{% extends "dtale/base.html" %}

{% block full_content %}
<input type="hidden" id="iframe" value={{iframe}} />
<div class="container-fluid">
    <div id="content">
    </div>
</div>
{% endblock %}

{% block js %}
<script type="text/javascript" src="{{ url_for('static', filename='dist/dtale_bundle.js') }}"></script>
{% endblock %}
